/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/execution/AbstractShellExecutor.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.execution;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang.time.StopWatch;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * There are four (4) different ways to run a ShellCommand. Which one you use depends on your needs:
 * 
 * <ul>
 *   <li>executeInteractive - returns the ProcessContext to the call. You can wrap this in a ShellResult by hand or use the next method
 *   <li>executeAsync - returns a ShellResult for the command. Call waitFor() to wait until the script completes. This is useful if
 *                      you want to retrieve the shell stdout/stderr even if the script is terminated due to running too long (see waitFor(long)).
 *   <li>execute - Without the timeout, runs the shell command and waits for the default timeout for a script to finish (1 year)
 *   <li>execute - With the timeout, runs the shell command and waits for the script to finish within the specified number of milliseconds
 * </ul>
 *
 */
abstract class AbstractShellExecutor implements ShellExecutor {

    private static final NALogger logger = NALogger.getLogger(AbstractShellExecutor.class);
    
    /** The env vars that were set by SMO and not inherited from the parent process **/
    protected List<EnvironmentVariable> nonInheritedEnvVariables;
    
    private ProcessBuilder processBuilder; 
    private File workingDirectory;
    private ShellCommandTransformer shellCommandTransformer;
    
    protected AbstractShellExecutor() {
        this.processBuilder = new ProcessBuilder();
        this.nonInheritedEnvVariables = new LinkedList<EnvironmentVariable>();
    }
    
    protected AbstractShellExecutor(ShellCommandTransformer shellCommandTransformer) {
        this();
        this.shellCommandTransformer = shellCommandTransformer;
    }

    public abstract ProcessContext executeInteractive(ShellRunnable shellCommand);
    
    public final ShellResult execute(ShellRunnable shellCommand) {
        return new ShellResult(executeInteractive(shellCommand)).waitFor();
    }
    
    public final ShellResult execute(ShellRunnable shellCommand, long timeoutInMsec) {
        return new ShellResult(executeInteractive(shellCommand)).waitFor(timeoutInMsec);
    }
    
    public final ShellResult executeAsync(ShellRunnable shellCommand) {
        return new ShellResult(executeInteractive(shellCommand));
    }
    
    /**
     * For the UnixUserSwitchingShell we have to write out the contents of the environment and the 
     * command to execute into a file because the users environment could wipe out part of the
     * environment that was setup to run the command.
     * 
     * su user_name -c "child_command" "argument a" "argument b" "argument c" 
     * 
     * @see com.netapp.common.execution.AbstractShellExecutor#doExecute(com.netapp.common.execution.ShellRunnable)
     */
    protected ProcessContext doExecute(ShellRunnable shellCommand) {
        ShellRunnable shellCommandToRun = (shellCommandTransformer != null) ? shellCommandTransformer.transform(shellCommand) : shellCommand;
        try {
            logger.debug(MsgKey.EXECUTION_EXECUTING_SHELL_COMMAND, shellCommandToRun.getLogFacingCommand());
            
            File directory = (shellCommandToRun.getWorkingDirectory() != null ? shellCommandToRun.getWorkingDirectory() : this.workingDirectory);
            
            if (directory != null) {
                this.processBuilder.directory(directory);
            }
            
            this.processBuilder.command(shellCommandToRun.getArgumentList());
            StopWatch timer = new StopWatch();
            timer.start();
            return new ProcessContext(shellCommandToRun, this.processBuilder.start(), timer);
        }
        catch (IOException e) {
            // TODO log
            throw new RuntimeException(e);
        }
    }
    
    public void setEnvironmentVariable(EnvironmentVariable environmentVariable) {
        this.processBuilder.environment().put(environmentVariable.getKey(), environmentVariable.getValue());
        
        // If it was set programatically then it was not inherited by the parent process of server process 
        this.nonInheritedEnvVariables.add(environmentVariable);
    }

    public void unsetEnvironmentVariable(String key) {
        this.processBuilder.environment().remove(key);
        
        for (Iterator<EnvironmentVariable> iterator = this.nonInheritedEnvVariables.iterator(); iterator.hasNext();) {
            EnvironmentVariable next = iterator.next();
            if (next.getKey().equals(key)) {
                iterator.remove();
            }
        }
    }
   
    /**
     * Gets the value of the given environment variable name. If this variable has not 
     * been programatically set by SMO it will return the value of the env variable that 
     * the SMO server was started with.
     * 
     * @see com.netapp.common.execution.ShellExecutor#getEnvironmentVariable(java.lang.String)
     */
    public EnvironmentVariable getEnvironmentVariable(String key) {
        return new EnvironmentVariable(key, this.processBuilder.environment().get(key));
    }
   
    /**
     * Returns the environment variables that were programatically set and not
     * inherited by the current process.
     * 
     * @return Environment variables set in the current context
     */
    public List<EnvironmentVariable> getNonInheritedEnvironmentVariables() {
        return this.nonInheritedEnvVariables;
    }
    
    public List<EnvironmentVariable> getUserRelevantEnvironmentVariables() {
        return new LinkedList<EnvironmentVariable>();
    }
    
    protected ProcessBuilder getProcessBuilder() {
        return this.processBuilder;
    }

    public File getWorkingDirectory() {
        return workingDirectory;
    }

    /**
     * Sets the working directory for the command that is going to be executed.  This value may be 
     * overridden by each ShellCommand.
     * @param workingDirectory The working directory.
     */
    public void setWorkingDirectory(File workingDirectory) {
        this.workingDirectory = workingDirectory;
    }

    ShellCommandTransformer getShellCommandTransformer() {
        return this.shellCommandTransformer;
    }

    void setShellCommandTransformer(ShellCommandTransformer shellCommandTransformer) {
        this.shellCommandTransformer = shellCommandTransformer;
    }
    
    public String getUser() {
        return System.getProperty("user.name");
    }
}
