/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/TaskInstanceTemplate.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.*;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.w3c.dom.Node;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.util.StringUtils;
import com.netapp.common.util.Utils;
import com.netapp.common.xml.DOMUtils;

class TaskInstanceTemplate implements ExecutionStep {
    private static final NALogger logger = NALogger.getLogger(TaskInstanceTemplate.class);

    private final StackFrameTemplate stackFrame;
    private Class<? extends Task> task;
    private String displayName;
    private FailureStrategy failureStrategy = FailureStrategy.IMMEDIATE;

    private TaskGuard taskGuard;
    
    /**
     * Our input parameters. Key is the name of the field
     * in the Task definition. 
     */
    private Set<InputParameter> inputs = new HashSet<InputParameter>();

    /**
     * Our output parameters. Key is the output register
     */
    private Set<OutputParameter> outputs = new HashSet<OutputParameter>();

    private TaskInstanceFactory taskInstanceFactory;

    TaskInstanceTemplate(StackFrameTemplate stackFrameTemplate, Node xml, TaskInstanceFactory taskInstanceFactory) {
        this.stackFrame = stackFrameTemplate;
        this.taskInstanceFactory = taskInstanceFactory;

        this.displayName = xml.getLocalName();
        String className = DOMUtils.getAttribute(xml, CLASSNAME_ATTRIBUTE);
        if (StringUtils.isBlank(className)) {
            throw new FlowException(logger.error(MsgKey.FLOW_CONFIGURATION_TASK_MISSING_CLASSNAME, xml.getNodeName()));
        }

        try {
            Class<?> tmpClass = Class.forName(className);
            if (Task.class.isAssignableFrom(tmpClass)) {
                this.task = tmpClass.asSubclass(Task.class);
            }
            else {
                throw new FlowException(logger.error(MsgKey.FLOW_TASK_CLASS_NOT_TASK, xml.getNodeName(), className));
            }
            // Check to make sure we have a legal constructor
            tmpClass.getConstructor();

        }
        catch (ClassNotFoundException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_TASK_CLASS_NOT_FOUND, xml.getNodeName(), className), e);
        }
        catch (SecurityException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_CONFIGURATION_ERROR, e), e);
        }
        catch (NoSuchMethodException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_TASK_CLASS_MISSING_CONSTRUCTOR, xml.getNodeName(),
                className), e);
        }

        this.failureStrategy = FailureStrategy.valueOf(DOMUtils.getAttribute(xml, FAILURE_STRATEGY_ATTRIBUTE));

        for (Node child = xml.getFirstChild(); child != null; child = child.getNextSibling()) {
            if (DOMUtils.isElement(child, WORKFLOW_SCHEMA_NAMESPACE, INPUT_ELEMENT)) {
                addInput(child);
            }
            else if (DOMUtils.isElement(child, WORKFLOW_SCHEMA_NAMESPACE, OUTPUT_ELEMENT)) {
                addOutput(child);
            }
        }
        
        this.initializeGuard(xml);
    }

    protected void setFailureStrategy(FailureStrategy strategy) {
        this.failureStrategy = strategy;
    }

    private void initializeGuard(Node xml) {
        String ifRegisterName = DOMUtils.getAttribute(xml, IF_CONDITION_ATTRIBUTE);
        String unlessRegisterName = DOMUtils.getAttribute(xml, UNLESS_CONDITION_ATTRIBUTE);

        if (ifRegisterName == null && unlessRegisterName == null) {
            this.taskGuard = TRUE_TASK_GUARD;
            return;
        }
        
        if (ifRegisterName != null && unlessRegisterName != null) {
            throw new FlowException(logger.error(MsgKey.FLOW_TASK_SPECIFIES_BOTH_IF_AND_UNLESS_GUARDS, xml.getNodeName()));
        }

        String conditionRegisterName = ifRegisterName != null ? ifRegisterName : unlessRegisterName;
        RegisterTemplate register = this.stackFrame.getRegisterInstanceTemplate(conditionRegisterName);
        
        if (register == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_TASK_GUARD_REGISTER_NOT_DEFINED, conditionRegisterName, this.task
                .getName()));
        }

        boolean negate = unlessRegisterName != null;
        this.taskGuard = new RegisterTaskGuard(register, negate);
    }
    
    @SuppressWarnings("unchecked")
    private void addInput(Node node) {
        String name = DOMUtils.getAttribute(node, NAME_ATTRIBUTE);
        List<Method> setters = ClassUtils.findSetterMethod(this.task, name);

        InputParameter param = null;

        String inputValue = node.getTextContent().trim();

        if (StringUtils.isEmpty(inputValue)) {
            Method[] path = null;
            String[] registerPath = StringUtils.coalesce(DOMUtils.getAttribute(node, REGISTER_ATTRIBUTE), name).split(
                "\\.");
            String registerName = registerPath[0];

            RegisterTemplate register = this.stackFrame.getRegisterInstanceTemplate(registerName);
            if (register == null) {
                throw new FlowException(logger.error(MsgKey.FLOW_INPUT_REGISTER_NOT_DEFINED, registerName, this.task
                    .getName()));
            }

            Type registerType = register.getType();

            if (registerPath.length > 1) {
                path = new Method[registerPath.length - 1];
                for (int i = 1; i < registerPath.length; i++) {
                    Class<?> clazz = ClassUtils.extractImmediateClass(registerType);
                    Method m = ClassUtils.findGetterMethod(clazz, registerPath[i]);
                    if (m == null) {
                        throw new FlowException(logger.error(MsgKey.FLOW_INPUT_PROPERTY_NOT_FOUND, registerPath[i],
                            registerName));
                    }
                    registerType = ClassUtils.extractFieldTypeFromBeanMethod(m);
                    path[i - 1] = m;
                }
            }

            Method setter = findSetterForType(setters, registerType);
            if (setter == null) {
                throw new FlowException(logger.error(MsgKey.FLOW_INPUT_SETTER_NOT_FOUND, this.task.getName(), name,
                    registerType));
            }
            
            Class<?> setterClass = setter.getParameterTypes()[0];
            
            if (ClassUtils.typeAssignableFromCollection(setterClass)) {
                param = getExistingInputParam(this.inputs, setter);
                
                if (null == param) {
                    param = new CollectionInputRegisterParameter(setter, setterClass, path);
                }

                ((CollectionInputRegisterParameter) param).addRegister(register);
            }
            else {
                param = new InputRegisterParameter(setter, register, path);
            }

            logger.internal("Mapping input " + name + " => " + register.toString());
        }
        else {
            if (setters.size() != 1) {
                throw new FlowException(logger.error(MsgKey.FLOW_INPUT_DEFAULT_SETTER_NOT_FOUND, this.task.getName(),
                    name));
            }
            /*
             * If this input has already been added and is an array type, pull the existing value
             * and add to it
             */
            param = getExistingInputParam(this.inputs, setters.get(0));
            Class<?> clazz = setters.get(0).getParameterTypes()[0];

            if (null == param) {
                Object inputObject = null;
                
                if (ClassUtils.typeAssignableFromCollection(clazz)) {
                    inputObject = ClassUtils.createCollection(clazz, inputValue);
                }
                else {
                    inputObject = ClassUtils.createObject(clazz, inputValue);
                }
                
                param = new InputValueParameter(setters.get(0), inputObject);
                logger.internal("Setting InputValueParameter method " + param.getSetter().toString() + " with object "
                    + inputObject);
            }
            else {
                if (null == param.getValue(null)) {
                    logger.internal("Parameter " + param.getSetter().toString() + " has a null value");
                }
                /*
                 * Taking out support for straight arrays for now as they do not work.
                 * 
                else if (param.getValue(null).getClass().isArray()) {
                    // TODO: As it stands, this will not work without more modification.
                    //       The array is stored in the param object, but we cannot update
                    //       the value property. The value is only exposed via a getter.
                    //       This means we cannot actually exist the array and store it back
                    //       into the inputs
                    int length = Array.getLength(param.getValue(null));
                    //Object newArray = Array.newInstance(param.getValue(null).getClass(), length + 1);
                    Array.set(param.getValue(null), length, inputValue);
                    logger.internal("Adding input " + name + " to existing array");
                }
                */
                else if (ClassUtils.typeAssignableFromCollection(param.getValue(null).getClass())) {
                    Collection<Object> temp = (Collection<Object>) param.getValue(null);
                    logger.internal("Adding input " + name + " to existing list: " + param.getValue(null));
                    temp.add(inputValue);
                }
                else {
                    logger.internal("Failed to map input " + name + " with value " + inputValue);
                }
            }

            logger.internal("Mapping input " + name + " => \"" + inputValue + "\"");
        }

        if (null != param) {
            this.inputs.add(param);
        }
    }

    private void addOutput(Node node) {
        String name = DOMUtils.getAttribute(node, NAME_ATTRIBUTE);
        String registerName = StringUtils.coalesce(DOMUtils.getAttribute(node, REGISTER_ATTRIBUTE), name);
        boolean append = Boolean.valueOf(DOMUtils.getAttribute(node, APPEND_ATTRIBUTE));

        RegisterTemplate register = this.stackFrame.getRegisterInstanceTemplate(registerName);
        if (register == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_OUTPUT_REGISTER_NOT_DEFINED, registerName, this.task
                .getName()));
        }

        Method getter = ClassUtils.findGetterMethod(this.task, name);
        
        TaskOutput taskOutput = getter == null ? null : getter.getAnnotation(TaskOutput.class);
        boolean downcastAllowed = taskOutput != null && taskOutput.downcastable();
        
        OutputParameter param = null;
        
        if (getter != null) {
            if (append) {
                /*
                 * For the append case, the register must be a collection and the
                 * object returned from the getter must be able to be inserted
                 * into the collection. 
                 */
                
                if (ClassUtils.isGenericCollection(register.getType())
                    && (downcastAllowed || ClassUtils.typeAssignableFrom(
                        ((ParameterizedType) register.getType()).getActualTypeArguments()[0],
                        getter.getGenericReturnType()))) {
                        
                    param = new AppendOutputParameter(getter, register);
                }
            }
            else {
                if ((downcastAllowed || ClassUtils.typeAssignableFrom(
                    register.getType(), 
                    getter.getGenericReturnType()))) {
                    
                    param = new BaseOutputParameter(getter, register);
                }
            }
        }
        
        if (param == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_OUTPUT_GETTER_NOT_FOUND, this.task.getName(), name,
                register.getType()));
        }
        outputs.add(param);
        logger.internal("Mapping output " + name + " => " + register.toString());
    }

    private Method findSetterForType(List<Method> setters, Type fieldType) {
        for (Method m : setters) {
            Type methodType = m.getGenericParameterTypes()[0];

            if (ClassUtils.typeAssignableFrom(methodType, fieldType)) {
                return m;
            }

            Class<?> methodClass = m.getParameterTypes()[0];

            if (ClassUtils.typeAssignableFromCollection(methodClass)) {
                return m;
            }
        }
        return null;
    }

    public void execute(StackFrame stackFrame) throws Throwable {
        
        if (!this.taskGuard.evaluate(stackFrame)) {
            logger.internal("Skipping task %s.", this.task.getName());
            return;
        }
        
        logger.internal("Executing task %s", this.task.getName());

        try {
            Task taskInstance = this.taskInstanceFactory.newTaskInstance(this.task, stackFrame.getContext());
            for (InputParameter input : inputs) {
                input.setInput(taskInstance, stackFrame);
            }

            taskInstance.execute();

            for (OutputParameter output : outputs) {
                output.setOutput(taskInstance, stackFrame);
            }
        }
        catch (InstantiationException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_OPERATION_TASK_REFLECTION_EXCEPTION, e, getDisplayName()), e);
        }
        catch (IllegalAccessException e) {
            throw new FlowException(logger.error(MsgKey.FLOW_OPERATION_TASK_REFLECTION_EXCEPTION, e, getDisplayName()), e);
        }
        catch (RetryableException e) {
            // Bubble this up
            throw e;
        }
        // CHECKSTYLE:OFF
        catch (Throwable t) {
            // CHECKSTYLE:ON
            stackFrame.getContext().setFailureStrategy(
                FailureStrategy.max(stackFrame.getContext().getFailureStrategy(), this.failureStrategy));

            if (this.failureStrategy == FailureStrategy.IGNORE) {
                logger.warn(MsgKey.FLOW_OPERATION_TASK_FAILURE_IGNORED, t, 
                    getDisplayName(), 
                    t.getLocalizedMessage());
            }
            else if (this.failureStrategy == FailureStrategy.CHECKPOINT || this.failureStrategy == FailureStrategy.END) {
                logger.error(MsgKey.FLOW_OPERATION_TASK_FAILURE_DELAYED, t, 
                    getDisplayName(),
                    t.getLocalizedMessage());
            }
            else if (this.failureStrategy == FailureStrategy.IMMEDIATE) {
                stackFrame.getContext().setFailureStrategy(FailureStrategy.IMMEDIATE);
                logger.error(MsgKey.FLOW_OPERATION_TASK_FAILURE, t, getDisplayName(), Utils.prettyString(t));
                throw t;
            }
        }
    }

    private InputParameter getExistingInputParam(Set<InputParameter> inputs, Method method) {
        InputParameter parameter = null;

        for (InputParameter input : inputs) {
            if (input.getSetter().equals(method)) {
                parameter = input;
            }
        }

        return parameter;
    }

    /**
     * Interface for input parameters. This interface defines a method to retrieve
     * the current value when this task is instantiated.
     * @author sunday
     * @version $Revision: #1 $ $Date: 2010/08/20 $
     */
    private abstract class InputParameter {
        protected Method setter;

        InputParameter(Method setter) {
            this.setter = setter;
        }

        public Method getSetter() {
            return this.setter;
        }

        public void setInput(Task taskInstance, StackFrame context) {
            Exception ex = null;
            try {
                this.setter.invoke(taskInstance, this.getValue(context));
            }
            catch (IllegalArgumentException e) {
                ex = e;
            }
            catch (IllegalAccessException e) {
                ex = e;
            }
            catch (InvocationTargetException e) {
                ex = e;
            }
            if (ex != null) {
                throw new FlowException(logger.error(MsgKey.FLOW_INPUT_REFLECTION_EXCEPTION, ex, this.setter,
                    taskInstance.getClass()), ex);
            }
        }

        protected abstract Object getValue(StackFrame context);

        @Override
        public int hashCode() {
            return this.setter.hashCode();
        }
    }

    private class InputValueParameter extends InputParameter {
        private Object obj;

        InputValueParameter(Method setter, Object obj) {
            super(setter);
            this.obj = obj;
        }

        @Override
        protected Object getValue(StackFrame context) {
            return this.obj;
        }
    }

    private class InputRegisterParameter extends InputParameter {
        private RegisterTemplate register;
        protected Method[] propertyPath;

        InputRegisterParameter(Method setter, RegisterTemplate register, Method... propertyPath) {
            super(setter);
            this.register = register;
            this.propertyPath = propertyPath;
        }

        @Override
        protected Object getValue(StackFrame context) {
            Object obj = context.getRegisterValue(this.register.getName());
            return getValueFromPropertyPath(obj);
        }
        
        protected Object getValueFromPropertyPath(Object value) {
            if (this.propertyPath != null) {
                for (Method m : this.propertyPath) {
                    Exception ex = null;
                    try {
                        value = m.invoke(value);
                    }
                    catch (IllegalArgumentException e) {
                        ex = e;
                    }
                    catch (IllegalAccessException e) {
                        ex = e;
                    }
                    catch (InvocationTargetException e) {
                        ex = e;
                    }
                    if (ex != null) {
                        throw new FlowException(logger.error(MsgKey.FLOW_INPUT_PROPERTY_REFLECTION_EXCEPTION, m
                            .getName(), value.getClass().getName()));
                    }
                }
            }
            
            return value;
        }
    }
    
    private class CollectionInputRegisterParameter extends InputRegisterParameter {
        private Collection<RegisterTemplate> registers = new ArrayList<RegisterTemplate>();
        private Class<?> setterClass;
        
        CollectionInputRegisterParameter(Method setter, Class<?> setterClass, Method... propertyPath) {
            super(setter, null, propertyPath);
            
            this.setterClass = setterClass;
        }
        
        protected void addRegister(RegisterTemplate register) {
            this.registers.add(register);
        }
        
        @SuppressWarnings("unchecked")
        @Override
        protected Object getValue(StackFrame context) {
            Collection<Object> collection = (Collection<Object>) ClassUtils.createCollection(this.setterClass, null);
            
            for (RegisterTemplate register : this.registers) {
                Object obj = context.getRegisterValue(register.getName());
                obj = getValueFromPropertyPath(obj);
                
                if (obj != null) {
                    if (ClassUtils.typeAssignableFromCollection(obj.getClass())) {
                        collection.addAll((Collection<Object>) obj);
                    }
                    else {
                        collection.add(obj);
                    }
                }
            }

            return collection;
        }
    }
    
    private abstract class OutputParameter {
        protected Method getter;
        protected RegisterTemplate register;

        OutputParameter(Method getter, RegisterTemplate register) {
            this.getter = getter;
            this.register = register;
        }

        abstract void setOutput(Task taskInstance, StackFrame context);

        @Override
        public int hashCode() {
            return register.hashCode();
        }
    }

    private class BaseOutputParameter extends OutputParameter {

        BaseOutputParameter(Method getter, RegisterTemplate register) {
            super(getter, register);
        }

        @Override
        public void setOutput(Task taskInstance, StackFrame context) {
            Exception ex = null;
            try {
                Object obj = getter.invoke(taskInstance);
                context.setRegisterValue(this.register.getName(), obj);
            }
            catch (IllegalArgumentException e) {
                ex = e;
            }
            catch (IllegalAccessException e) {
                ex = e;
            }
            catch (InvocationTargetException e) {
                ex = e;
            }
            if (ex != null) {
                throw new FlowException(logger.error(MsgKey.FLOW_OUTPUT_REFLECTION_EXCEPTION, ex, this.getter,
                    taskInstance.getClass()), ex);
            }
        }
    }

    private class AppendOutputParameter extends OutputParameter {
        AppendOutputParameter(Method getter, RegisterTemplate register) {
            super(getter, register);
        }

        @Override
        @SuppressWarnings("unchecked")
        // Cast to Collection<Object>
        void setOutput(Task taskInstance, StackFrame context) {
            Object obj;
            Exception ex = null;
            try {
                obj = getter.invoke(taskInstance);
                Collection<Object> collection = (Collection<Object>) context.getRegisterValue(this.register.getName());
                if (collection == null) {
                    // Initialize the register with the proper Collection type 
                    collection = (Collection<Object>) ClassUtils.createCollection(ClassUtils.extractImmediateClass(this.register.getType()), null);
                    context.setRegisterValue(this.register.getName(), collection);
                }
                
                if (ClassUtils.typeAssignableFromCollection(obj.getClass())) {
                    collection.addAll((Collection<Object>) obj);
                }
                else {
                    collection.add(obj);
                }
            }
            catch (IllegalArgumentException e) {
                ex = e;
            }
            catch (IllegalAccessException e) {
                ex = e;
            }
            catch (InvocationTargetException e) {
                ex = e;
            }
            if (ex != null) {
                throw new FlowException(logger.error(MsgKey.FLOW_OUTPUT_REFLECTION_EXCEPTION, ex, this.getter,
                    taskInstance.getClass()), ex);
            }
        }
    }

    /**
     * Get the display name of this task. Currently, this is the name of the
     * node used in the XML declaration. Errors that can occur during normal
     * operation should use this name to display to the user.
     * @return Display name of the task
     */
    public String getDisplayName() {
        // TODO: Add a way to localize this through an annotation on the
        // Task class or XML definition.
        return this.displayName;
    }
    
    @Override
    public String toString() {
        return this.task.getName();
    }

    private static interface TaskGuard {
        boolean evaluate(StackFrame stackFrame);
    }

    private static final TaskGuard TRUE_TASK_GUARD = new TaskGuard() {
        public boolean evaluate(StackFrame stackFrame) {
            return true;
        }
    };
    
    private class RegisterTaskGuard implements TaskGuard {
        
        private RegisterTemplate register;
        private boolean negate;

        private RegisterTaskGuard(RegisterTemplate register, boolean negate) {
            this.register = register;
            this.negate = negate;
        }
        
        public boolean evaluate(StackFrame stackFrame) {
            boolean result = this.negate ^ (Boolean) stackFrame.getRegisterValue(this.register.getName());
            
            logger.internal("Task %s guard condition %s%s evaluates to %b.", 
                TaskInstanceTemplate.this.task.getName(),
                this.negate ? "!" : "",
                this.register,
                result
            );
            
            return result;
        }
    }
}