/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/generator/WorkflowSchemaAntTask.java#1 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */

package com.netapp.common.flow.generator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.FileSet;

/**
 * TaskSchemaGenerator is an AntTask implemented to support the generation 
 * of Nnf Schema catalog XSD files for a set of implementing Task 
 * classes within a project.
 * 
 * Only concrete classes which implement the necessary interface or extend 
 * the necessary class are generated into the schema file.
 * 
 * To include within Ant, first a task def for the Nnf ant tools needs to 
 * be included in the build .xml
 * Note: {SLASH} used rather than / when would have cause a end comment :-)
 * 
 * <taskdef resource="nnfAntTasks.properties">
 *   <classpath>
 *     <!-- the fileset for the NnfAntTools jar and dependencies -->
 *     <fileset dir="/u/dholmes/p4/NNFPlus/dist">
 *       <include name="**{SLASH}*.jar"/>
 *    </fileset>
 *     <!-- the compiled files we need to scan -->
 *     <pathelement location="${buildclassesdir}"/>
 *     <!-- the other libraries that the compiled classes additionally depend on -->
 *     <fileset dir="${libdir}">
 *       <include name="**{SLASH}*.jar"/>
 *     </fileset>
 *   </classpath>
 * </taskdef>
 *
 * Then a target would be defined to use the nnf-catalog task.  Classes to 
 * be scanned are provided via a fileset, which would allow the separation 
 * of group of tasks to different schemas if desired by the using project.
 * 
 *   <target name="catalog" depends="compile">
 *     <workflow-schema
 *       schemaFileToGenerate="/u/dholmes/p4/armadaProto/testagentschema.xsd" 
 *       catalogSchemaNamespace="http://www.netapp.com/armada/flow"
 *   >
 *       <fileset dir="${buildclassesdir}">
 *         <patternset>
 *           <include name="**{SLASH}*.class"/>
 *         </patternset>
 *       </fileset>
 *     </workflow-schema>
 *  </target>
 *
 * @author dholmes
 */
public class WorkflowSchemaAntTask extends Task {
    private Vector<FileSet> filesets = new Vector<FileSet>();
    private File schemaFileToGenerate;
    private String interfaceRequiredToImplement = NNF_DEFAULT_TASK_INTERFACE;
    private String classRequiredToExtend;
    private String catalogSchemaNamespace;
    private boolean stripTaskSuffix;
    private boolean useFullyQualifiedClassName;

    public static final String NNF_DEFAULT_TASK_INTERFACE = "com.netapp.common.flow.Task";

    public WorkflowSchemaAntTask() {
    }

    public void addFileset(FileSet set) {
        this.filesets.add(set);
    }

    /**
     * Namespace schema URL to use for the generated namespace
     * @param catalogSchemaNamespace
     */
    public void setCatalogSchemaNamespace(String catalogSchemaNamespace) {
        this.catalogSchemaNamespace = catalogSchemaNamespace;
    }

    /**
     * Tasks included in the schema must implement an interface or extend a particular project class.
     * This interface is optional and defaults to the NNF core Task interface
     * @param interfaceRequiredToImplement
     */
    public void setInterfaceRequiredToImplement(String interfaceRequiredToImplement) {
        this.interfaceRequiredToImplement = interfaceRequiredToImplement;
    }

    /**
     * Filepath for the output schema file to generate
     * @param schemaFileToGenerate
     */
    public void setSchemaFileToGenerate(File schemaFileToGenerate) {
        this.schemaFileToGenerate = schemaFileToGenerate;
    }

    /**
     * Tasks included in the schema must implement an interface or extend a particular project class.
     * This item is optional and may only apply for a project which may include certain abstract base 
     * classes for some reason.
     * @param classRequiredToExtend
     */
    public void setClassRequiredToExtend(String classRequiredToExtend) {
        this.classRequiredToExtend = classRequiredToExtend;
    }

    /**
     * Strip the "Task" suffix from the classname when generating the XSD element name?
     */
    public void setStripTaskSuffix(boolean stripTaskSuffix) {
        this.stripTaskSuffix = stripTaskSuffix;
    }
    
    /**
     * Use the fully qualified classname when generating the XSD element name?
     */
    public void setUseFullyQualifiedClassName(boolean useFullyQualifiedClassName) {
        this.useFullyQualifiedClassName = useFullyQualifiedClassName;
    }
    
    public void execute() {
        try {
            // get the files from the fileset
            // adapted from ant Copy task
            List<String> filesToProcess = new ArrayList<String>();
            for (int i = 0; i < this.filesets.size(); i++) {
                FileSet fs = this.filesets.elementAt(i);
                DirectoryScanner ds = fs.getDirectoryScanner(getProject());
                filesToProcess.addAll(Arrays.asList(ds.getIncludedFiles()));
            }

            this.generateFromClassFiles(filesToProcess);
        }
        // CHECKSTYLE:OFF
        catch (Exception e) {
        // CHECKSTYLE:ON
            throw new BuildException(e);
        }

    }

    private void generateFromClassFiles(List<String> javaClassFiles) throws IOException, ClassNotFoundException {

        if (this.schemaFileToGenerate.exists()) {
            System.out.println("Schema file " + this.schemaFileToGenerate.getCanonicalPath() + " exists.  Overwriting ..");
        }

        this.schemaFileToGenerate.delete();

        PrintWriter schemaWriter = new PrintWriter(new FileWriter(this.schemaFileToGenerate));

        schemaWriter.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        schemaWriter.println("<xs:schema");
        schemaWriter.println("   xmlns:xs=\"http://www.w3.org/2001/XMLSchema\"");
        schemaWriter.println("   xmlns:fc=\"http://www.netapp.com/common/flow/workflow\"");
        schemaWriter.println("   xmlns=\"" + this.catalogSchemaNamespace + "\"");
        schemaWriter.println("   targetNamespace=\"" + this.catalogSchemaNamespace + "\"");
        schemaWriter.println("   elementFormDefault=\"qualified\"");
        schemaWriter.println(">");
        schemaWriter.println("<xs:import namespace=\"http://www.netapp.com/common/flow/workflow\"/>");
        schemaWriter.println();

        try {
            Class<?> interfaceClass = (this.interfaceRequiredToImplement != null) ? Class.forName(this.interfaceRequiredToImplement) : null;
            Class<?> extendClass = (this.classRequiredToExtend != null) ? Class.forName(this.classRequiredToExtend) : null;


            for (String javaClassFile : javaClassFiles) {
                // Need to get the Class object
                String properClassName = javaClassFile.replace(File.separatorChar, '.');
                properClassName = properClassName.endsWith(".class") ? properClassName.substring(0, properClassName.length() - 6) : properClassName;
                Class<?> c = Class.forName(properClassName, false, this.getClass().getClassLoader());

                // Make the decision
                if (!Modifier.isAbstract(c.getModifiers())
                        && !Modifier.isInterface(c.getModifiers())
                        && ((interfaceClass != null && interfaceClass.isAssignableFrom(c))
                                || (extendClass != null && extendClass.isAssignableFrom(c)))) {
                    
                    String fullPathClassName = c.getName();
                    
                    String className = this.useFullyQualifiedClassName ? fullPathClassName : c.getSimpleName();
                    
                    if (this.stripTaskSuffix) {
                        if (className.endsWith("Task")) {
                            className = className.substring(0, className.length() - "Task".length());
                        }
                    }
                    
                    schemaWriter.println("<xs:element name=\"" + className + "\" substitutionGroup=\"fc:Task\">");
                    schemaWriter.println("  <xs:complexType>");
                    schemaWriter.println("    <xs:complexContent>");
                    schemaWriter.println("      <xs:extension base=\"fc:taskType\">");
                    schemaWriter.println("        <xs:attribute name=\"classname\" type=\"xs:string\" fixed=\"" + fullPathClassName + "\"/>");
                    schemaWriter.println("      </xs:extension>");
                    schemaWriter.println("    </xs:complexContent>");
                    schemaWriter.println("  </xs:complexType>");
                    schemaWriter.println("</xs:element>");
                    schemaWriter.println();
                }
            }
        }
        finally {
            schemaWriter.println("</xs:schema>");
            schemaWriter.close();
        }

    }


}
