/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/windows/WMIConnection.java#1 $
 *
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved. 
 */
package com.netapp.common.platform.windows;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Methods to query Windows Management Instrumentation.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public abstract class WMIConnection {
    private static final NALogger logger = NALogger.getLogger(WMIConnection.class);
    
    protected final String namespace;
    
    /**
     * Construct a WMI object for the selected namespace.
     * @param namespace
     */
    protected WMIConnection(String namespace) {
        this.namespace = namespace;
    }
    
    /**
     * Create a new connection to WMI
     * @param namespace the WMI namespace to connect to
     * @return WMI connection to the specified namespace
     * @throws UnsupportedOperationException The WMI native code could not be loaded or WMI is not supported on this platform.
     */
    public static WMIConnection getConnection(String namespace) {
        try {
            return new WMINativeImpl(namespace);
        }
        catch (UnsatisfiedLinkError e) {
            throw new UnsupportedOperationException(
                logger.debug(MsgKey.PLATFORM_WMI_METHOD_DEPENDS_ON_NATIVE_LIBRARY, e), e);
        }
    }
    
    /**
     * Execute a WQL query and marshall the results back into map.
     * @param query WQL query to execute
     * @return
     */
    public abstract Collection<Map<String, Object>> query(String query);

    /**
     * Get a WMI object by path
     * @return Map for the specified object
     */
    public abstract Map<String, Object> getObject(String path);
    
    /**
     * Execute a method on an object and marshall the results back into
     * a map.
     * @param path
     * @param method
     * @param inParams
     * @return
     */
    public abstract Map<String, Object> execMethod(String path, String method, Map<String, Object> inParams);
    
    /**
     * Execute a WQL query and marshall the results back into a collection of WMI
     * object proxies.
     * @param clazz Interface extending WbemObject for the resultant objects
     * @param query
     * @return Collection of T proxies
     */
    public <T extends WbemObject> Collection<T> query(Class<T> clazz, String query) {
        Collection<Map<String, Object>> queryResults = query(query);
        Collection<T> results = null;
        if (queryResults != null) {
            results = new ArrayList<T>();
            for (Map<String, Object> map : queryResults) {
                results.add(proxyClass(clazz, map));
            }
        }
        return results;
    }
    
    /**
     * Get a WMI object as a proxy class
     * @param clazz
     * @param path
     * @return WMI object
     */
    public <T extends WbemObject> T getObject(Class<T> clazz, String path) {
        return proxyClass(clazz, getObject(path));
    }
    
    /**
     * Shorthand method when one object is expected for query(). If multiple values are
     * returned, the object returned may be indeterminate.
     * @param query
     * @return query result object
     */
    public Map<String, Object> queryForObject(String query) {
        Collection<Map<String, Object>> collection = query(query);
        if (collection == null) {
            return null;
        }
        assert collection.size() == 1;
        for (Map<String, Object> obj : collection) {
            return obj;
        }
        return null;
    }
    
    /**
     * Shorthand method when one object is expected for query(). If multiple values are
     * returned, the object returned may be indeterminate.
     * @param clazz
     * @param query
     * @return query result object
     */
    public <T extends WbemObject> T queryForObject(Class<T> clazz, String query) {
        return proxyClass(clazz, queryForObject(query));
    }
    
    /**
     * Query for the associators of a given object
     * @param path
     * @param resultClass
     * @return Collection of associators
     */
    public Collection<Map<String, Object>> associatorsOf(String path, String resultClass) {
        String query = "ASSOCIATORS OF {" + path + "}";
        if (resultClass != null) {
            query += " WHERE ResultClass=" + resultClass;
        }
        return query(query);
    }
    
    /**
     * Query for the associators of a given object and marshall the answers
     * into a WMI proxy object
     * @param clazz
     * @param path
     * @param resultClass
     * @return Collection of associators
     */
    public <T extends WbemObject> Collection<T> associatorsOf(Class<T> clazz, String path, String resultClass) {
        Collection<Map<String, Object>> queryResults = associatorsOf(path, resultClass);
        Collection<T> results = null;
        if (queryResults != null) {
            results = new ArrayList<T>();
            for (Map<String, Object> map : queryResults) {
                results.add(proxyClass(clazz, map));
            }
        }
        return results;
    }
    
    /**
     * Query for the associators of a given object and marshall the
     * answers into a WMI proxy object.
     * @param <T>
     * @param clazz
     * @param object
     * @param resultClass
     * @return
     */
    public <T extends WbemObject> Collection<T> associatorsOf(Class<T> clazz, WbemObject object, String resultClass) {
        return associatorsOf(clazz, object.getRelpath(), resultClass);
    }
    
    /**
     * Construct a proxy class over a result map.
     * Getter methods annotated with {@link WMIElement} will return
     * the indicated value from the underlying map.
     * Methods annotated with {@link WMIMethod} and {@link WMIParameter}
     * will invoke the named method on the underlying WMI object and return a Map
     * with the output parameters.
     * @param <T>
     * @param clazz
     * @param result
     * @return
     */
    public <T extends WbemObject> T proxyClass(Class<T> clazz, Map<String, Object>result) {
        if (result != null) {
            return clazz.cast(Proxy.newProxyInstance(clazz.getClassLoader(), new Class[] {clazz}, new WMIResultProxy(result)));
        }
        return null;
    }
    
    private class WMIResultProxy implements InvocationHandler {
        private Map<String, Object> map;
        
        public WMIResultProxy(Map<String, Object> map) {
            this.map = map;
        }
        
        // TODO: This is not particularly robust at the moment.
        public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
            if (method.getDeclaringClass() == Object.class || method.getDeclaringClass() == Map.class) {
                if (method.getName().equals("finalize")) {
                    return null;
                }
                if (method.getName().equals("toString") && method.getParameterTypes().length == 0) {
                    return this.toString();
                }
                return method.invoke(map, args);
            }
            WMIElement element = method.getAnnotation(WMIElement.class);
            if (element != null) {
                return this.map.get(element.value());
            }
            
            WMIMethod wmiMethod = method.getAnnotation(WMIMethod.class);
            if (wmiMethod != null) {
                Map<String, Object> in = new HashMap<String, Object>();
                Annotation[][] parameters = method.getParameterAnnotations();
                for (int i = 0; i < args.length; i++) {
                    for (Annotation a : parameters[i]) {
                        if (a instanceof WMIParameter) {
                            in.put(((WMIParameter)a).value(), args[i]);
                        }
                    }
                }
                return WMIConnection.this.execMethod((String)this.map.get("__RELPATH"), wmiMethod.value(), in);
            } 
            return null;
        }
        
        @Override
        public String toString() {
            List<String> keys = new ArrayList<String>(this.map.keySet());
            Collections.sort(keys);
            StringBuilder b = new StringBuilder();
            b.append("{\n");
            for (String key : keys) {
                Object value = this.map.get(key);
                if (value instanceof Object[]) {
                    value = Arrays.deepToString((Object[])value);
                }
                b.append(String.format("  %-25s ==> %s\n", key, value));
            }
            b.append("}");
            return b.toString();
        }
    }
    
    public static void main(String[] args) {
        String namespace = "root\\cimv2";
        int next = 0;
        if (args.length == 0) {
            System.err.println("Usage: WMIConnection [namespace] query");
            System.exit(1);
        }
        if (args.length > 1) {
            namespace = args[next];
            next++;
        }
        
        String query = args[next];
        System.out.println("Query " + namespace + " " + query);
        WMIConnection wmi = WMIConnection.getConnection(namespace);
        Collection<Map<String, Object>> results = wmi.query(query);
        for (Map<String, Object> map : results) {
            System.out.println("{");
            for (Map.Entry<String, Object> entry : map.entrySet()) {
                System.out.println("    " + entry.getKey() + "    " + entry.getValue());
            }
            System.out.println("}");
            System.out.println();
        }
    }
}
