/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/TypedefData.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.netapp.common.MsgKey;
import com.netapp.common.zapi.annotation.Element;
import com.netapp.common.zapi.annotation.Optional;
import com.netapp.common.zapi.annotation.Typedef;

/**
 * Represents the reflective properties of a zapi typedef struct.
 * A zapi typedef represents a grouping of elements used for
 * input and output in a Zephyr api call.
 * 
 * @see Typedef
 * @see Element
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
final class TypedefData {
    /**
     * Cache of already processed Typedef objects keyed on the java class
     */
    private static Map<Class<?>, TypedefData> typedefs = new ConcurrentHashMap<Class<?>, TypedefData>();
    
    /**
     * The java class object containing this typedef struct
     */
    protected Class<?> typedefClass;
    
    /**
     * The zapi element name of this typedef.
     */
    protected String zapiTypedefName;
    
    /**
     * The collection of element fields in this typedef.
     * Key is the element's zapi name
     */
    protected Map<String, FieldData> zapiElements;
    
    /**
     * Construct a new reflective map for the supplied class by looking
     * for the attached {@link Element} annotations on fields or methods.
     * Clients should use {@link #getTypedefData(Class)} instead of this.
     * @param javaClass Java class containing a definition for this typedef.
     * @throws ZAPIRunnerException an error was detected with the class or field annotations in the supplied class object.
     */
    private TypedefData(Class<?> javaClass) {
        this.typedefClass = javaClass;
        Typedef typedef = this.typedefClass.getAnnotation(Typedef.class);
        if (typedef == null) {
            throw new ZAPIRunnerException(ZAPIRunner.logger.error(MsgKey.ZAPI_RUNNER_CLASS_NOT_ZAPI_TYPEDEF, this.typedefClass));
        }
        
        if (typedef.value().length() > 0) {
            this.zapiTypedefName = typedef.value();
        }
        else {
            this.zapiTypedefName = ZAPIRunner.javaToZapi(this.typedefClass.getSimpleName());
        }
        
        zapiElements = new HashMap<String, FieldData>();

        // Check for annotations on the getters and setters
        Method[] methods = this.typedefClass.getDeclaredMethods();
        for (Method m : methods) {
            boolean optional = m.isAnnotationPresent(Optional.class);
            Element element = m.getAnnotation(Element.class);
            if (element != null) {
                FieldData fieldData = new FieldData(m, element.value(), optional);
                if (!this.zapiElements.containsKey(fieldData.getZapiFieldName())) {
                    this.zapiElements.put(fieldData.getZapiFieldName(), fieldData);
                }
                else {
                    throw new ZAPIRunnerException(ZAPIRunner.logger.error(
                        MsgKey.ZAPI_RUNNER_ELEMENT_DECLARED_TWICE, 
                        fieldData.getZapiFieldName(), 
                        this.typedefClass));
                }
            }
        }
    }

    /**
     * Get the java class associated with this typedef
     * @return
     */
    public Class<?> getTypedefClass() {
        return this.typedefClass;
    }
    
    /**
     * Get the declared zapi name of this typedef
     * @return
     */
    public String getZapiTypedefName() {
        return this.zapiTypedefName;
    }
    
    /**
     * Get a map of the declared field elements on this typedef.
     * @return Map of elements keyed by zapi typedef name
     */
    public Map<String, FieldData> getZapiElements() {
        return this.zapiElements;
    }
    
    /**
     * Retrieve the TypedefData class from cache or construct a new
     * one for the supplied object. This is a convienence function
     * equivilent to {@link #getTypedefData(Class) getTypedefData(obj.getClass())}
     * 
     * @param obj Object to retrieve the typedef data for
     * @return typedef data for the class of obj
     */
    static TypedefData getTypedefData(Object obj) {
        return getTypedefData(obj.getClass());
    }
    
    /**
     * Retrieve the TypedefData class from cache or construct a new
     * one for the supplied class.
     * @param clazz Class to retrieve the typedef data for
     * @return typedef data for the class
     */
    static TypedefData getTypedefData(Class<?> clazz) {
        TypedefData typedefData = typedefs.get(clazz);
        if (typedefData == null) {
            typedefData = new TypedefData(clazz);
            typedefs.put(clazz, typedefData);
        }
        return typedefData;
    }
}