/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/generator/wsdl/WsdlGenerator.java#1 $
 *
 * Copyright (c) 2009 NetApp, Inc.
 * All rights reserved. 
 */
package com.netapp.common.zapi.generator.wsdl;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.commons.lang.ClassUtils;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import com.jamesmurty.utils.XMLBuilder;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.CollectionUtils;
import com.netapp.common.util.Transformer;
import com.netapp.common.zapi.annotation.StatusEnum;
import com.netapp.common.zapi.generator.CategoryGenerator;
import com.netapp.common.zapi.parser.Api;
import com.netapp.common.zapi.parser.Category;
import com.netapp.common.zapi.parser.Errdef;
import com.netapp.common.zapi.parser.Field;
import com.netapp.common.zapi.parser.Namespace;
import com.netapp.common.zapi.parser.Type;
import com.netapp.common.zapi.parser.Typedef;
import com.netapp.common.zapi.parser.ZapidocParser;

/**
 * Generates a WSDL from zapidoc.
 * 
 * One WSDL is generated with the schema inline.
 * 
 * 
 * @author sunday
 * @author philipc
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class WsdlGenerator {
    
    static final NALogger logger = NALogger.getLogger(WsdlGenerator.class);

    private final File outputFile;
    private final Namespace zapiNamespace;
    private final Map<String, Errdef> errdefs;
    private final Set<String> errnosNotFound;
    private final String xmlNamespace; 
    private final String endpoint;
    private final Collection<StatusEnum> apiStatusesToInclude;

    private XMLBuilder wsdlBuilder;
    private List<XMLBuilder> messageBuilders;
    private XMLBuilder portTypeBuilder;
    private XMLBuilder bindingBuilder;
    private Map<String, String> zapiTypesNeedingArrayWrappers = new TreeMap<String, String>();
    private Map<String, String> zapiTypesNeedingWrappers = new TreeMap<String, String>();
    private XMLBuilder schemaBuilder;

    private WsdlGeneratorFactory generatorFactory;
    
    private static final Map<BigIntegerRange, String> INTEGER_TYPES;
    static {
        Map<BigIntegerRange, String> integerTypes = new HashMap<BigIntegerRange, String>();
        
        integerTypes.put(new BigIntegerRange(Byte.toString(Byte.MIN_VALUE), Byte.toString(Byte.MAX_VALUE)), "xsd:byte");
        integerTypes.put(new BigIntegerRange("0", "255"), "xsd:unsignedByte");
        integerTypes.put(new BigIntegerRange(Short.toString(Short.MIN_VALUE), Short.toString(Short.MAX_VALUE)), "xsd:short");
        integerTypes.put(new BigIntegerRange("0", "65535"), "xsd:unsignedShort");
        integerTypes.put(new BigIntegerRange(Integer.toString(Integer.MIN_VALUE), Integer.toString(Integer.MAX_VALUE)), "xsd:int");
        integerTypes.put(new BigIntegerRange("0", "4294967295"), "xsd:unsignedInt");
        integerTypes.put(new BigIntegerRange(Long.toString(Long.MIN_VALUE), Long.toString(Long.MAX_VALUE)), "xsd:long");
        integerTypes.put(new BigIntegerRange("0", "18446744073709551615"), "xsd:unsignedLong");
        
        INTEGER_TYPES = Collections.unmodifiableMap(integerTypes);
    }
    
    private static final Map<String, String> SIMPLE_TYPES;
    static {
        
        Map<String, String> simpleTypes = new HashMap<String, String>();
        // Types allowed by spec
        simpleTypes.put("string", "xsd:string");
        simpleTypes.put("boolean", "xsd:boolean");
        simpleTypes.put("integer", "xsd:integer");
        
        // DFM
        simpleTypes.put("bool", "xsd:boolean");
        
        // Currently, obj-type is not defined by dfm zapidoc in 3.7, so we assume it's just an alias for a string
        // this is fixed in 3.8+
//        simpleTypes.put("obj-type", "xsd:string");
        
        // TODO: FIX - A typedef of empty needs a container for attributes
        simpleTypes.put("empty", "xsd:string");
        SIMPLE_TYPES = Collections.unmodifiableMap(simpleTypes);
    }
    
    private static class BigIntegerRange {
        private final BigInteger min;
        private final BigInteger max;
        public BigIntegerRange(String min, String max) {
            this(new BigInteger(min), new BigInteger(max));
        }
        public BigIntegerRange(BigInteger min, BigInteger max) {
            super();
            this.min = min;
            this.max = max;
        }
        public BigInteger getMin() {
            return this.min;
        }
        
        public BigInteger getMax() {
            return this.max;
        }
        
        @Override
        public boolean equals(Object obj) {
            return EqualsBuilder.reflectionEquals(this, obj);
        }
        
        @Override
        public int hashCode() {
            return HashCodeBuilder.reflectionHashCode(this);
        }
        
    }
    
    private WsdlGenerator(File outputFile, ZapidocParser parser, String zapiNamespace, String xmlNamespace, String endpoint, Collection<StatusEnum> apiStatusesToInclude) {

        this.outputFile = outputFile;
        this.zapiNamespace = parser.getNamespace(zapiNamespace);
        this.xmlNamespace = xmlNamespace;
        this.endpoint = endpoint;
        this.apiStatusesToInclude = apiStatusesToInclude;
        
        this.errdefs = parser.getErrdefsMap();
        
        this.errnosNotFound = new HashSet<String>();
        
    }
    
    public void generateWsdl() throws IOException, ParserConfigurationException, TransformerException {
        
        this.generatorFactory = new WsdlGeneratorFactory();
        this.generatorFactory.addWordReplacements(Arrays.asList(this.zapiNamespace));

        generate();
        
        write();
    }

    private void generate() throws ParserConfigurationException, IOException, TransformerException {
        
        this.schemaBuilder = XMLBuilder.create("xsd:schema")
            .a("targetNamespace", WsdlGenerator.this.xmlNamespace)
            .a("attributeFormDefault", "unqualified")
            .a("elementFormDefault", "qualified");
        
        this.messageBuilders = new ArrayList<XMLBuilder>();
        
        String namespaceGenerated = WsdlGenerator.this.generatorFactory.getNamespaceGenerator(WsdlGenerator.this.zapiNamespace).getNameGenerated();
        this.portTypeBuilder = XMLBuilder.create("portType")
            .a("name", namespaceGenerated  + "Interface");

        appendCategoryDocumentations();
        
        this.bindingBuilder = XMLBuilder.create("binding")
            .a("name", namespaceGenerated + "Binding")
            .a("type", "na:" + namespaceGenerated + "Interface")
            .e("soap:binding")
                .a("style", "document")
                .a("transport", "http://schemas.xmlsoap.org/soap/http")
            .up();
        
        Set<Type> referencedTypes = new TreeSet<Type>();
        
        /*
         * Go through all the APIs (so their input/output elements appear at the top) 
         */
        for (Type t : WsdlGenerator.this.zapiNamespace.getTypes()) {
            if (t instanceof Api) {
                Api api = (Api) t;
                if (shouldGenerate(api)) {
                    appendApi(api);
                    referencedTypes.addAll(api.getAllReferencedTypes());
                }
                    
            }
        }
        
        /*
         * Through all the other referenced typedefs  
         */
        for (Type t : referencedTypes) {
            if (t instanceof Typedef) {
                appendTypedef((Typedef) t);
            }
        }
        
        /*
         * Append all the wrappers (gathered during the first two steps) 
         */
        
        for (Map.Entry<String, String> typeToWrap : this.zapiTypesNeedingWrappers.entrySet()) {
            appendWrapper(typeToWrap.getKey(), typeToWrap.getValue(), false);
        }
        
        for (Map.Entry<String, String> typeToWrap : this.zapiTypesNeedingArrayWrappers.entrySet()) {
            appendWrapper(typeToWrap.getKey(), typeToWrap.getValue(), true);
        }
        
        /*
         * Append some standard types
         */
        appendStandardTypes();
        
        
        /*
         * Now put everything together
         */

        this.wsdlBuilder = XMLBuilder.create("definitions")
            .a("name", "NetApp" + namespaceGenerated)
            .a("targetNamespace", WsdlGenerator.this.xmlNamespace)
            .a("xmlns:na", WsdlGenerator.this.xmlNamespace)
            .a("xmlns:wsi", "http://ws-i.org/schemas/conformanceClaim/")
            .a("xmlns:xsd", "http://www.w3.org/2001/XMLSchema")
            .a("xmlns:soap", "http://schemas.xmlsoap.org/wsdl/soap/")
            .a("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
            .a("xmlns", "http://schemas.xmlsoap.org/wsdl/");
        
        
        XMLBuilder typesBuilder = this.wsdlBuilder
            .e("types");
        
        
        typesBuilder.getElement().appendChild(
            this.wsdlBuilder.getDocument().importNode(this.schemaBuilder.getElement(), true)
        );
        
        for (XMLBuilder messageBuilder : this.messageBuilders) {
            
            this.wsdlBuilder.getElement().appendChild(
                this.wsdlBuilder.getDocument().importNode(messageBuilder.getElement(), true)
            );
        }
        
        this.wsdlBuilder.getElement().appendChild(
            this.wsdlBuilder.getDocument().importNode(this.portTypeBuilder.getElement(), true)
        );
        
        
        this.wsdlBuilder.getElement().appendChild(
            this.wsdlBuilder.getDocument().importNode(this.bindingBuilder.getElement(), true)
        );
        

        this.wsdlBuilder
            .e("service")
            .a("name", namespaceGenerated + "Service")
                .e("port")
                .a("name", namespaceGenerated  + "Port")
                .a("binding", "na:" + namespaceGenerated + "Binding")
                    .e("soap:address")
                    .a("location", WsdlGenerator.this.endpoint);

    }

    private void appendCategoryDocumentations() throws ParserConfigurationException {
        Set<Category> categoriesToDocument = new TreeSet<Category>();
        for (Type t : WsdlGenerator.this.zapiNamespace.getTypes()) {
            if (t instanceof Api) {
                Api api = (Api) t;
                if (shouldGenerate(api)) {
                    categoriesToDocument.add(api.getCategory());
                }
            }
        }
        
        XMLBuilder portTypeDocumentationBuilder = this.portTypeBuilder.e("documentation");
        for (Category category : categoriesToDocument) {
            appendCategoryDocumentation(portTypeDocumentationBuilder, category);
        }
    }

    private boolean shouldGenerate(Api api) {
        return WsdlGenerator.this.apiStatusesToInclude.contains(api.getStatus());
    }

    private void appendApi(Api t) throws ParserConfigurationException {
        
        WsdlTypeGenerator typeGenerator = WsdlGenerator.this.generatorFactory.getTypeGenerator(t);
        String nameGenerated = typeGenerator.getNameGenerated();
        String descGenerated = typeGenerator.getDescGenerated();
        
        /*
         * Input and Output elements
         */
        appendElement(nameGenerated, null, descGenerated, t.getInputFields());
        
        /*
         * The ZAPI response actually extends a result object that contains status, errno, and reason.
         * We use SOAP Faults to pass on failures, so we don't need to extend the common results object.
         * 
         * If we were to extend a common results object, use a line like the following.
         * appendElement(nameGenerated + "Result", "na:results", null, t.getOutputFields());
         */
        appendElement(nameGenerated + "Result", null, null, t.getOutputFields());
        
        /*
         * Input and Output messages (that reference the input and output elements)
         */
        this.messageBuilders.add(XMLBuilder.create("message")
            .a("name", nameGenerated + "Request")
                .e("part")
                    .a("name", "parameters")
                    .a("element", "na:" + nameGenerated)
                .up()); // end part
        
        this.messageBuilders.add(XMLBuilder.create("message")
            .a("name", nameGenerated + "Response")
                .e("part")
                    .a("name", "results")
                    .a("element", "na:" + nameGenerated + "Result")
                .up()); // end part
        
        /*
         * portType/operation (that reference the input and output messages)
         */
        XMLBuilder operationBuilder = this.portTypeBuilder
            .e("operation")
                .a("name", nameGenerated);
        
        appendOperationDocumentation(operationBuilder.e("documentation"), descGenerated, t.getStatus(), t.getCategory(), t.getErrors());
        
        operationBuilder
            .e("input")
                .a("message", "na:" + nameGenerated + "Request")
            .up() // end input
            .e("output")
                .a("message", "na:" + nameGenerated + "Response");
                
                    
        /*
         * binding/operation
         */
        this.bindingBuilder
            .e("operation")
                .a("name", nameGenerated)
                .e("soap:operation")
                    .a("soapAction", "urn:" + nameGenerated)
                .up() // end soap:operation
                .e("input")
                    .e("soap:body")
                        .a("use", "literal")
                    .up() //end soap:body
                .up() // end input
                .e("output")
                    .e("soap:body")
                        .a("use", "literal");
        
    }

    private void appendOperationDocumentation(XMLBuilder documentationBuilder, String desc, StatusEnum status, Category category, Set<String> errors) throws ParserConfigurationException {
        if (desc != null || !errors.isEmpty() || status != null || category != null) {
            
            XMLBuilder operationDocumentationBuilder = documentationBuilder
                .e("OperationDocumentation")
                    .a("xsi:type", "na:OperationDocumentation");
            
            if (desc != null) {
                operationDocumentationBuilder
                    .e("na:Description")
                        .text(desc);
                
            }
            
            if (category != null) {
                operationDocumentationBuilder
                    .e("na:Category")
                        .text(this.generatorFactory.getCategoryGenerator(category).getNameGenerated());
            }
                
            if (status != null) {
                operationDocumentationBuilder
                    .e("na:Status")
                        .text(status.getValue());
            
            }
            
            if (!errors.isEmpty()) {
                
                XMLBuilder operationErrorsBuilder = operationDocumentationBuilder
                    .e("na:OperationErrors");
                
                for (String error : errors) {
                    
                    
                    XMLBuilder operationErrorBuilder = operationErrorsBuilder
                        .e("na:OperationError");
                    
                    Errdef errdef = WsdlGenerator.this.errdefs.get(error);
                    if (errdef != null) {
                        operationErrorBuilder
                            .e("na:Code")
                                .text(Integer.toString(errdef.getCode()));
                    }
                    else {
                        /*
                         * Warning already generated in ZapidocParser
                         */
                    }
                            
                    operationErrorBuilder
                        .e("na:Name")
                            .text(error);
                }
            }
        }
    }
    
    private void appendCategoryDocumentation(XMLBuilder documentationBuilder, Category category) throws ParserConfigurationException {
        if (category != null) {
            
            CategoryGenerator categoryGenerator = this.generatorFactory.getCategoryGenerator(category);
            
            XMLBuilder categoryDocumentationBuilder = documentationBuilder
                .e("CategoryDocumentation")
                    .a("xsi:type", "na:CategoryDocumentation");
            
            if (category.getName() != null) {
                categoryDocumentationBuilder
                    .e("na:Name")
                        .text(categoryGenerator.getNameGenerated());
            }
            
            if (category.getSynopsis() != null) {
                categoryDocumentationBuilder
                    .e("na:Synopsis")
                        .text(categoryGenerator.getSynopsisGenerated());
            }
            
            if (category.getDiscussion() != null) {
                categoryDocumentationBuilder
                    .e("na:Discussion")
                        .text(categoryGenerator.getDiscussionGenerated());
            }
        }
    }
    
    
    /**
     * Appends an element as a child of the schemaBuilder
     */
    private void appendElement(String name, String baseType, String desc, Collection<Field> fields) {
        
        appendType(
            this.schemaBuilder
                .e("xsd:element")
                .a("name", name),
            null, baseType, null, desc, fields);
    }
    
    private void appendTypedef(Typedef t) {
        WsdlTypeGenerator typeGenerator = WsdlGenerator.this.generatorFactory.getTypeGenerator(t);
        
        appendType(this.schemaBuilder, 
            typeGenerator.getNameGenerated(), 
            null, 
            t.getType(), 
            typeGenerator.getDescGenerated(), 
            t.getFields());
    }
    
    /**
     * Appends a complexType as a child of the parentBuilder (which is either a schema or an element)
     */
    private void appendType(XMLBuilder parentBuilder, String typeName, String baseTypeName, String aliasedTypeName, String desc, Collection<Field> fields) {
        
        if (aliasedTypeName != null) {
            
            
            Type aliasedType = WsdlGenerator.this.zapiNamespace.getType(aliasedTypeName);
            
            if (aliasedType.getRootType().isSimple()) {
                
                String base;
                if (aliasedType.isSimple()) {
                    base = SIMPLE_TYPES.get(aliasedTypeName);
                }
                else {
                    base = "na:" + WsdlGenerator.this.generatorFactory.getTypeGenerator(aliasedType).getNameGenerated();
                }
                
                XMLBuilder simpleTypeBuilder = parentBuilder
                    .e("xsd:simpleType")
                        .a("name", typeName);
            
                appendXsdAnnotationDocumentation(simpleTypeBuilder, desc);
            
                simpleTypeBuilder
                    .e("xsd:restriction")
                        .a("base", base);
                
            }
            else {
                String complexBase = "na:" + WsdlGenerator.this.generatorFactory.getTypeGenerator(aliasedType).getNameGenerated();
                /*
                 * Somebody in the inheritance chain is not simple.
                 * Therefore we have to add a complex content inheritance.
                 * 
                 * Since we already handle that below,
                 * just recursively call back into this method with the 
                 * baseTypeName switched with the aliasTypeName
                 */
                
                appendType(parentBuilder, typeName, complexBase, null, desc, fields);
            }
            
        }
        else {
            
            XMLBuilder complexTypeBuilder = parentBuilder
                .e("xsd:complexType");

            if (typeName != null) {
                /*
                 * Name will be null when we're appending an element whose parent is the schema.
                 * Name will be non-null when we're appending an element whose parent is another complexType
                 */
                complexTypeBuilder
                    .a("name", typeName);
                
            }
            
            appendXsdAnnotationDocumentation(complexTypeBuilder, desc);
            
            if (baseTypeName != null) {
                complexTypeBuilder = complexTypeBuilder
                    .e("xsd:complexContent")
                        .e("xsd:extension")
                            .a("base", baseTypeName);
            }
            
            XMLBuilder allBuilder = null;
            for (Field f : fields) {
                
                WsdlFieldGenerator fieldGenerator = WsdlGenerator.this.generatorFactory.getFieldGenerator(f);
                
                if (allBuilder == null) {
                    allBuilder = complexTypeBuilder
                        .e("xsd:all");
                }
                
                String simpleType = SIMPLE_TYPES.get(f.getType());

                /*
                 * ZAPI xml format is described here:
                 * http://wikid.netapp.com/w/HHG/Software-4.0/Zephyr_APIs-12.0/API_Guidelines
                 * 
                 * There are some weird semantics in the zapi guidelines.
                 * I've outlined the different cases below.
                 */
                
                XMLBuilder elementBuilder = allBuilder
                    .e("xsd:element")
                        .a("name", fieldGenerator.getNameGenerated());
                
                if (f.isOptional()) {
                    elementBuilder
                        .a("minOccurs", "0");
                }
                
                appendXsdAnnotationDocumentation(elementBuilder, fieldGenerator.getDescGenerated());
                
                if (!f.isCollection()) {
                    
                    if (simpleType != null) {
                        /*
                         * Case 1: Non-array, simpleType
                         * 
                         * Schema looks like:
                         * 
                         * <xsd:element name="${field.name}" type="${simpleType}"/>
                         * 
                         * XML looks like:
                         * 
                         * <${field.name}>value</${field.name}>
                         */
                        
                        
                        /*
                         * XXX HACK
                         * The api-proxy parameter request.args should be an xsd:any instead of string
                         * The api-proxy-results field response.results should be an xsd:any instead of string
                         * 
                         * The real fix is to introduce a new type into the ZAPIdoc that is mapped to xsd:any
                         */
                        if (f.getDeclaringType().getName().equals("request") && f.getName().equals("args")
                                || f.getDeclaringType().getName().equals("response") && f.getName().equals("results")) {
                            
                            assert simpleType.equals("xsd:string");

                            elementBuilder
                                .e("xsd:complexType")
                                    .e("xsd:sequence")
                                        .e("xsd:any")
                                            .a("minOccurs", "0")
                                            .a("maxOccurs", "unbounded");

                        }
                        else {
                            if (f.getRange() == null) {
                                
                                elementBuilder 
                                    .a("type", simpleType);
                                
                            }
                            else {
                                
                                String integerType = INTEGER_TYPES.get(new BigIntegerRange(f.getRange().getMinInteger(), f.getRange().getMaxInteger()));
                                
                                if (integerType != null) {
                                    elementBuilder 
                                        .a("type", integerType);
                                }
                                else {

                                    elementBuilder
                                        .e("xsd:simpleType")
                                            .e("xsd:restriction")
                                                .a("base", "xsd:integer")
                                                .e("xsd:minInclusive")
                                                    .a("value", f.getRange().getMinInteger().toString())
                                                .up() // end xsd:minInclusive
                                                .e("xsd:maxInclusive")
                                                    .a("value", f.getRange().getMaxInteger().toString());
                                }
                                
                            }
                        }
                    }
                    else if (f.getName().equals(f.getType()) || f.getTypedefType().isAlias()) {
                        /*
                         * Case 2: Non-array, complexType, (field.type == field.name OR field.type.isAlias)
                         * 
                         * Schema looks like
                         * 
                         * <xsd:element name="${field.name}" type="na:${field.type}"/>
                         * 
                         * XML looks like:
                         * 
                         * <${field.name}>
                         *     <${type.field1}></${type.field1}>
                         *     ... 
                         * </${field.name}>
                         */
                        
                        elementBuilder
                            .a("type", "na:" + fieldGenerator.getTypeGenerated());
                        
                    }
                    else {
                        
                        /*
                         * Case 3: Non-array, complexType, (field.type != field.name AND !field.type.isAlias)
                         * 
                         * Schema looks like
                         * 
                         * <xsd:element name="${field.name}" type="na:wrapper-of-${field.type}"/>
                         * 
                         * XML looks like:
                         * 
                         * <${field.name}>
                         *     <${field.type}>
                         *         <${type.field1}></${type.field1}>
                         *         ... 
                         *     </${field.type}>
                         * </${field.name}>
                         * 
                         */
                        
                        /*
                         * Put the wrapper type in the wrapped type's xsd.
                         * This makes it so we don't have duplicate definitions
                         * of the wrapper in each of the referring xsds.
                         * 
                         * If fieldTypeCategory == null, that means this is a simpleType.
                         * The wrappers of simple types are output in the common xsd (CommonXsdGenerator)
                         */
                        addZapiTypeNeedingWrapper(f.getType(), fieldGenerator.getTypeGenerated());
                        
                        elementBuilder
                            .a("type", "na:" + getWrapperTypeName(fieldGenerator.getTypeGenerated()));
                        
                    }
                }
                else {
                    
                    /*
                     * Case 4: array
                     * 
                     * Schema looks like
                     * 
                     * <xsd:element name="${field.name}" type="na:array-of-${field.type}"/>
                     * 
                     * XML looks like this for simple types
                     * 
                     * <${field.name}>
                     *     <${field.type}>value</${field.type}>
                     *     ... 
                     * </${field.name}>
                     * 
                     * XML looks like this for complex types
                     * 
                     * <${field.name}>
                     *     <${field.type}>
                     *         <${type.field1}></${type.field1}>
                     *         ... 
                     *     </${field.type}>
                     *     <${field.type}>
                     *         <${type.field1}></${type.field1}>
                     *         ... 
                     *     </${field.type}>
                     * </${field.name}>
                     * 
                     */
                    
                    /*
                     * Put the wrapper type in the wrapped type's xsd.
                     * This makes it so we don't have duplicate definitions
                     * of the wrapper in each of the referring xsds.
                     * 
                     * If fieldTypeCategory == null, that means this is a simpleType.
                     * The wrappers of simple types are output in the common xsd (CommonXsdGenerator)
                     */
                    addZapiTypeNeedingArrayWrapper(f.getType(), fieldGenerator.getTypeGenerated());

                    elementBuilder
                        .a("type", "na:" + getArrayWrapperTypeName(fieldGenerator.getTypeGenerated()));
                }
            }
        }
    }

    private void appendXsdAnnotationDocumentation(XMLBuilder builder, String desc) {
        if (desc != null) {
            builder
                .e("xsd:annotation")
                    .e("xsd:documentation")
                        .text(desc);
        }
    }
    
    private void appendWrapper(String zapiTypeToWrap, String transformedTypeToWrap, boolean isArray) {

        String type = SIMPLE_TYPES.get(zapiTypeToWrap);
        if (type == null) {
            type = "na:" + transformedTypeToWrap;
        }
        
        String wrapperTypeName = isArray
            ? getArrayWrapperTypeName(transformedTypeToWrap)
            : getWrapperTypeName(transformedTypeToWrap);
        
        XMLBuilder elementBuilder = this.schemaBuilder
            .e("xsd:complexType")
                .a("name", wrapperTypeName)
                .e("xsd:sequence")
                    .e("xsd:element")
                        .a("name", transformedTypeToWrap)
                        .a("type", type);
        
        if (isArray) {
            elementBuilder
                .a("maxOccurs", "unbounded");
        }
        
    }
    
    private String getWrapperTypeName(String typeToWrap) {
        return "WrapperOf" + typeToWrap;
    }
    
    private String getArrayWrapperTypeName(String typeToWrap) {
        return "ArrayOf" + typeToWrap;
    }
    
    private void addZapiTypeNeedingWrapper(String zapiType, String tranformedType) {
        this.zapiTypesNeedingWrappers.put(zapiType, tranformedType);
    }

    private void addZapiTypeNeedingArrayWrapper(String zapiType, String tranformedType) {
        this.zapiTypesNeedingArrayWrappers.put(zapiType, tranformedType);
    }
    
    private void appendStandardTypes() {
        /*
         * The netapp type is a wrapper used by zapi (but not in the soap interface).
         * It is commented out here because it is not used in soap.
         * 
        this.schemaBuilder
            .e("xsd:element")
                .a("name", "netapp")
                .e("xsd:complexType")
                    .e("xsd:sequence")
                        .e("xsd:any")
                            .a("minOccurs", "0")
                            .a("maxOccurs", "unbounded")
                        .up()
                    .up()
                    .e("xsd:attribute")
                        .a("name", "version")
                        .a("type", "xsd:string");
         */

        /*
         * The results type is the base type for all result objects.
         * 
         * It is commented out here because it is not used in soap.
         * (Errors are returned via SOAP Faults)
         * 
        this.schemaBuilder
            .e("xsd:complexType")
                .a("name", "results")
                .e("xsd:attribute")
                    .a("name", "reason")
                    .a("type", "xsd:string")
                .up()
                .e("xsd:attribute")
                    .a("name", "status")
                    .a("type", "na:resultsStatus")
                .up()
                .e("xsd:attribute")
                    .a("name", "errno")
                    .e("xsd:simpleType")
                        .e("xsd:restriction")
                            .a("base", "xsd:integer")
                            .e("xsd:minInclusive")
                                .a("value", "0")
                            .up()
                            .e("xsd:maxInclusive")
                                .a("value", "2147483647");
        
        this.schemaBuilder
            .e("xsd:simpleType")
                .a("name", "resultsStatus")
                .e("xsd:restriction")
                    .a("base", "xsd:string")
                    .e("xsd:enumeration")
                        .a("value", "passed")
                    .up()
                    .e("xsd:enumeration")
                        .a("value", "failed");
         */
        
        appendFaultDetail();
        appendOperationErrorType();
        appendArrayOfOperationErrorType();
        appendOperationDocumentationType();
        appendOperationStatusType();
        appendCategoryDocumentationType();
    }

    private void appendFaultDetail() {
        XMLBuilder faultDetailTypeBuilder = this.schemaBuilder
            .e("xsd:complexType");
        
        appendXsdAnnotationDocumentation(faultDetailTypeBuilder,
            "Faults will use this type to provide access to error information and reason.");
        
        faultDetailTypeBuilder
            .a("name", "FaultDetail")
            .e("xsd:sequence")
                .e("xsd:element")
                    .a("name", "OperationError")
                    .a("type", "na:OperationError")
                .up()
                .e("xsd:element")
                    .a("name", "Reason")
                    .a("type", "xsd:string");
    }
    
    private void appendOperationErrorType() {
        XMLBuilder operationErrorTypeBuilder = this.schemaBuilder
            .e("xsd:complexType");
    
        appendXsdAnnotationDocumentation(operationErrorTypeBuilder,
            "OperationError is returned as part of FaultDetail, and used for documenting possible errors returned from an operation");
        
        operationErrorTypeBuilder
            .a("name", "OperationError")
            .e("xsd:sequence")
                .e("xsd:element")
                    .a("name", "Code")
                    .a("type", "xsd:unsignedInt")
                .up()
                .e("xsd:element")
                    .a("name", "Name")
                    .a("type", "xsd:string");
    }

    private void appendArrayOfOperationErrorType() {
        this.schemaBuilder
            .e("xsd:complexType")
                .a("name", "ArrayOfOperationError")
                .e("xsd:sequence")
                    .e("xsd:element")
                        .a("name", "OperationError")
                        .a("type", "na:OperationError")
                        .a("maxOccurs", "unbounded");
    }

    private void appendOperationDocumentationType() {
        XMLBuilder operationDocumentationTypeBuilder = this.schemaBuilder
            .e("xsd:complexType");
    
        appendXsdAnnotationDocumentation(operationDocumentationTypeBuilder,
            "The OperationDocumentation type is used to document operations.");
        
        operationDocumentationTypeBuilder
            .a("name", "OperationDocumentation")
            .e("xsd:sequence")
                .e("xsd:element")
                    .a("name", "Description")
                    .a("type", "xsd:string")
                    .a("minOccurs", "0")
                .up()
                .e("xsd:element")
                    .a("name", "Category")
                    .a("type", "xsd:string")
                    .a("minOccurs", "0")
                .up()
                .e("xsd:element")
                    .a("name", "Status")
                    .a("type", "na:OperationStatus")
                    .a("minOccurs", "0")
                .up()
                .e("xsd:element")
                    .a("name", "OperationErrors")
                    .a("type", "na:ArrayOfOperationError")
                    .a("minOccurs", "0");
    }

    private void appendOperationStatusType() {
        
        XMLBuilder operationStatusTypeBuilder = this.schemaBuilder
            .e("xsd:simpleType");
    
        XMLBuilder restrictionBuilder = operationStatusTypeBuilder
            .a("name", "OperationStatus")
            .e("xsd:restriction")
                .a("base", "xsd:string");
        

        for (StatusEnum status : StatusEnum.values()) {
            restrictionBuilder
                .e("xsd:enumeration")
                    .a("value", status.getValue());
            
        }
    }

    private void appendCategoryDocumentationType() {
        XMLBuilder categoryDocumentationTypeBuilder = this.schemaBuilder
            .e("xsd:complexType");
    
        appendXsdAnnotationDocumentation(categoryDocumentationTypeBuilder,
            "The CategoryDocumentation type is used to document categories.  Categories are groups of related operations.");
        
        categoryDocumentationTypeBuilder
            .a("name", "CategoryDocumentation")
            .e("xsd:sequence")
                .e("xsd:element")
                    .a("name", "Name")
                    .a("type", "xsd:string")
                .up()
                .e("xsd:element")
                    .a("name", "Synopsis")
                    .a("type", "xsd:string")
                    .a("minOccurs", "0")
                .up()
                .e("xsd:element")
                    .a("name", "Discussion")
                    .a("type", "xsd:string")
                    .a("minOccurs", "0");
    }

    private void write() throws FileNotFoundException, TransformerException {
        PrintWriter writer = new PrintWriter(this.outputFile);
        
        Properties outputProperties = new Properties();
        outputProperties.put(javax.xml.transform.OutputKeys.INDENT, "yes");        
        outputProperties.put("{http://xml.apache.org/xslt}indent-amount", "4");        
        this.wsdlBuilder.toWriter(writer, outputProperties);
            
        writer.close();
    }
    
    public static void main(String[] args) {
        if (args.length < 5) {
            printUsage();
            System.exit(1);
        }

        String outputFileName = args[0];
        
        /*
         * zapi namespace e.g. filer, dfm, or netcache
         */
        String zapiNamespace = args[1];
        
        /*
         * XML namespace.  e.g. https://www.netapp.com/management/v1 
         */
        String xmlNamespace = args[2];
        
        /*
         * Service endpoint.  e.g. https://HOST_NAME:8488/apis/soap/v1
         */
        String endpoint = args[3];
        
        /*
         * Statuses.
         */
        String[] statusNames = args[4].split(",");
        Collection<StatusEnum> statuses; 
        if (statusNames[0].equals("all")) {
            statuses = EnumSet.allOf(StatusEnum.class);
        }
        else {
            statuses = EnumSet.copyOf(CollectionUtils.transform(Arrays.asList(statusNames), new HashSet<StatusEnum>(), new Transformer<String, StatusEnum>() {
            
                public StatusEnum transform(String statusString) {
                    return StatusEnum.parse(statusString);
                }
            }));
        }
        
        List<File> zapidocFiles = CollectionUtils.transform(Arrays.asList(args).subList(5, args.length), new ArrayList<File>(), new Transformer<String, File>() {
        
            public File transform(String s) {
                return new File(s);
            }
        });
        
        try {
            File outputFile = new File(outputFileName);

            ZapidocParser parser = new ZapidocParser(zapidocFiles);
            
            if (!parser.getErrors().isEmpty()) {
                System.exit(1);
            }
            
            WsdlGenerator generator = new WsdlGenerator(outputFile, parser, zapiNamespace, xmlNamespace, endpoint, statuses);
            generator.generateWsdl();
        }
        catch (IOException e) {
            printUsage(e);
        }
        catch (ParserConfigurationException e) {
            printUsage(e);
        }
        catch (TransformerException e) {
            printUsage(e);
        }
    }
    
    /**
     * Print the usage of the main program
     *
     */
    static void printUsage() {
        System.err.println("Usage: " + ClassUtils.getShortClassName(WsdlGenerator.class) + " <output_file> <zapi_namespace> <xml_namespace> <endpoint> <status> <zapidoc_file> [<zapidoc_file> ...]");
        System.err.println("    <output_file>     = Output wsdl file name");
        System.err.println("    <zapi_namespace>  = The namespace in the zapidoc file for which to generate the WSDL");
        System.err.println("                        (e.g. dfm or filer)");
        System.err.println("    <xml_namespace>   = The xml namespace to use for the generated WSDL and Schema");
        System.err.println("                        (e.g. http://www.netapp.com/management/v1)");
        System.err.println("    <endpoint>        = The soap endopint to use in the WSDL");
        System.err.println("                        (e.g. https://HOST_NAME:8488/apis/soap/v1)");
        System.err.println("    <status>          = Comma separated list of api statuses to include in the WSDL");
        System.err.println("                        (or \"all\" to include every api)");
        System.err.println("    <zapidoc_file>    = Name of a zapidoc file to parse");
    }
    
    /**
     * Print the usage of the main program
     * @param t
     */
    static void printUsage(Throwable t) {
        printUsage();
        if (t != null) {
            t.printStackTrace(System.err);
        }
        System.exit(1);
    }

}
