/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/parser/ZapidocParser.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.parser;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Parse a supplied zapidoc file into a set of annotated Api and Typedef classes
 * for use with {@link com.netapp.common.zapi.ZAPIRunner}.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class ZapidocParser {
    static final NALogger logger = NALogger.getLogger(ZapidocParser.class);
    
    /** Input files containing the zapidocs */
    private List<File> docFiles;

    /** List of blocks read from the input file */
    List<Block> blocks;
    
    /** List of api categories declared */
    private Set<Category> categories;
    
    /** Current (last seen) category */
    private Category currentCategory;
    
    /**
     * Map of known namespace names and the corresponding namespace
     * TODO: Move this onto a Generator context instead of a static
     */
    private Map<String, Namespace> namespaces;
    
    /**
     * The last accessed namespace. Typedefs that don't have an explicit namespace
     * defined belong to the namespace of the last processed Api.
     * TODO: Move this to a Generator Context
     */
    private static Namespace lastNamespace;
    
    private Map<Integer, Errdef> errdefs = new TreeMap<Integer, Errdef>();
    
    private List<String> errors = new ArrayList<String>();
    private List<String> warnings = new ArrayList<String>();
    
    /**
     * Construct a new ZapiGenerator
     * @param packageName Root package for the generated classes
     * @param zapidocFiles Input zapidoc files
     * @throws IOException
     */
    public ZapidocParser(List<File> zapidocFiles) throws IOException {
        
        this.docFiles = zapidocFiles;
        this.categories = new TreeSet<Category>();
        this.namespaces = new HashMap<String, Namespace>();
        
        this.parseInput();
        
        this.checkErrnos();
        
    }

    Category getCurrentCategory() {
        return this.currentCategory;
    }
    
    /**
     * Retrieve the existing namespace by name or construct a new one.
     * @param name Name of the namespace to retrieve
     * @return Namespace with the given name
     */
    public Namespace getNamespace(String name) {
        name = name.toLowerCase();
        Namespace n = namespaces.get(name);
        if (n == null) {
            n = new Namespace(this, name);
            namespaces.put(name, n);
        }
        return n;
    }
    
    public Collection<Errdef> getErrdefs() {
        return this.errdefs.values();
    }
    
    public Map<String, Errdef> getErrdefsMap() {
        Map<String, Errdef> errdefsMap = new HashMap<String, Errdef>();
        for (Errdef errdef : getErrdefs()) {
            errdefsMap.put(errdef.getName(), errdef);
        }
        return errdefsMap;
    }
    
    /**
     * Return a collection of all declared namespaces
     * @return collection of namespaces
     */
    public Collection<Namespace> getNamespaces() {
        return namespaces.values();
    }
    
    /**
     * Set the last accessed namespace
     * @param namespace Last accessed namespace
     */
    void setLastNamespace(Namespace namespace) {
        lastNamespace = namespace;
    }
    
    /**
     * Retrieve the last namespace used
     * @return last used namespace
     */
    Namespace getLastNamespace() {
        if (lastNamespace == null) {
            lastNamespace = getNamespace("filer");
        }
        return lastNamespace;
    }
    
    /**
     * Parse the input file.
     * @throws IOException
     */
    private void parseInput() throws IOException {
        
        this.blocks = new ArrayList<Block>();
        
        for (File docFile : this.docFiles) {
            BufferedReader reader = null;
            
            try {
                reader = new BufferedReader(new FileReader(docFile));
                this.blocks.addAll(Block.readBlocks(reader));
            }
            finally {
                reader.close();
            }
        }
        
        for (Block block : this.blocks) {
            processBlock(block);
        }
    }

    /**
     * Process a block into zapidoc objects.
     * @param block
     */
    private void processBlock(Block block) {
        if (block.getTags().isEmpty()) {
            return;
        }
        // This is horrible.
        Queue<Tag> queue = block.getTags();
        
        Tag tag = queue.poll();
        while (tag != null) {
            logger.internal("Processing " + tag.getType());
            if (tag.is(TagEnum.CATEGORY)) {
                Category cat = new Category(tag.getFirstWord());
                cat.populate(queue);
                categories.add(cat);
                currentCategory = cat;
            }
            else if (tag.is(TagEnum.API)) {
                Api api = new Api(this, tag.getFirstWord());
                api.populate(queue);
            }
            else if (tag.is(TagEnum.TYPEDEF)) {
                Typedef typedef = new Typedef(this, tag.getFirstWord());
                typedef.populate(queue);
            }
            else if (tag.is(TagEnum.ERRDEF)) {
                String[] tmp = tag.getValue().split("\\s+");
                Errdef errdef = new Errdef(tmp[0], Integer.parseInt(tmp[1]));
                errdef.populate(queue);
                
                Errdef existingErrdef = this.errdefs.get(errdef.getCode());
                
                if (existingErrdef != null) {
                    appendWarning(MsgKey.ZAPI_GENERATOR_DUPLICATE_ERRDEF, errdef.getCode(), existingErrdef.getName(), errdef.getName());
                }
                else {
                    errdefs.put(errdef.getCode(), errdef);
                }
            }
            else {
                break;
            }
            tag = queue.poll();
        }
    }
    
    /*
     * Logs errors if no errdef is found for any of the errnos
     */
    private void checkErrnos() {
        
        Map<String, Errdef> errdefsMap = getErrdefsMap();
        
        for (Namespace namespace : getNamespaces()) {
            for (Type type : namespace.getTypes()) {
                if (type instanceof Api) {
                    Api api = (Api) type;
                    for (String errno : api.getErrors()) {
                        if (!errdefsMap.containsKey(errno)) {
                            appendError(MsgKey.ZAPI_GENERATOR_UNKNOWN_ERRNO, errno, api.getName());
                        }
                    }
                }
            }
        }
    }
    
    protected void appendError(MsgKey messageKey, Object... params) {
        this.errors.add(logger.error(messageKey, params));
    }
    
    protected void appendWarning(MsgKey messageKey, Object... params) {
        this.warnings.add(logger.warn(messageKey, params));
    }
    
    public List<String> getErrors() {
        return Collections.unmodifiableList(this.errors);
    }
    
    public List<String> getWarnings() {
        return Collections.unmodifiableList(this.errors);
    }
    
}
